import warnings
import functools
from hashlib import sha512

import numpy as np
import pandas as pd
import scipy.stats as st
from statsmodels.stats.multitest import multipletests

from PIL import Image
import matplotlib as mpl
import matplotlib.pyplot as plt
import seaborn as sns
from seaborn.categorical import _SwarmPlotter
import palettable

__all__ = [
    'data_to_random_seed', 'desaturate_color', 'brighten_color',
    'permutation_p', 'permutation_p_fast', 'multiple_comparisons_with_correction',
    'p2str', 'multiplot', 'plot_pvals_pairwise', 'make_canvas',
    'plot_faceness_to_resp', 'custom_scatter', 'plot_metric_to_fsi'
]


mpl_params = {
    'figure.figsize': (2,1.5), 'figure.dpi': 200,
    'lines.linewidth': 1,
    'axes.linewidth': 0.6, 'axes.labelpad': 1,
    'xtick.major.width': 0.6, 'xtick.major.size': 2, 'xtick.major.pad': 2,
    'xtick.minor.width': 0.5, 'xtick.minor.size': 1.5, 'xtick.minor.pad': 1.5,
    'ytick.major.width': 0.6, 'ytick.major.size': 2, 'ytick.major.pad': 2,
    'ytick.minor.width': 0.5, 'ytick.minor.size': 1.5, 'ytick.minor.pad': 1.5,
    'axes.spines.right': False, 'axes.spines.top': False,
    'axes.titlesize': 7, 'axes.labelsize': 7,
    'xtick.labelsize': 6, 'ytick.labelsize': 6,
    'legend.fontsize': 7, 'legend.title_fontsize': 7,
    'legend.handlelength': 1.5, 'legend.handletextpad': 0.5,
    'font.family': 'sans-serif',  'font.sans-serif': ['Arial'],
    'font.size': 7, 'svg.fonttype': 'none', 'legend.frameon': False,
    'savefig.dpi': 300, 'savefig.bbox': 'tight', 'savefig.facecolor': (1,1,1,0), 
}

colors1 = palettable.cartocolors.qualitative.Bold_10.mpl_colors
colors2 = palettable.cartocolors.qualitative.Pastel_10.mpl_colors
colors3 = palettable.scientific.sequential.Imola_9.mpl_colors
palette = {
    'Face': colors1[2],
    'Stylized face': colors2[5],
    'Non-face': colors1[3],
    'Stylized non-face': colors2[1],
    'CIT, high FSI': colors1[0],
    'Any region, low FSI': colors1[1],
    'Evolved, FSI > 1/3': colors1[-2],
    'Evolved, FSI < 0': colors1[-3]}
tal_categories = [
    'Human face', 'Human face w. PPE', 'Human face & body',
    'Monkey face', 'Monkey face & body', 'Other animate',
    'Inanimate face-like', 'Non-face']
palette2 = {k: v for k, v in zip(tal_categories, colors3)}
palette3 = palette.copy()
palette3.update(palette2)

num_permutations = 100000


def data_to_random_seed(*arrays):
    m = sha512()
    for a in arrays:
        m.update(a.tobytes())
    return int(m.hexdigest()[-4:], 16)


def desaturate_color(rgb, lam=0.5):
    return mpl.colors.hsv_to_rgb(np.array(mpl.colors.rgb_to_hsv(rgb)) * [1, 1-lam, 1])


def brighten_color(rgb, lam=0.7):
    return mpl.colors.hsv_to_rgb(np.array(mpl.colors.rgb_to_hsv(rgb)) * [1, 1, 1-lam] + [0, 0, lam])


def permutation_p(y1, y0, n_perm=num_permutations, alternative='greater', paired=False, statistic=np.mean, random_seed=0, range_=range):
    """  y0, y1: data for 2 conditions (0 and 1) respectively
         statstic: a function that calculates one number to describe data (y)
             range_: set to tqdm.trange to display progress bar
    """
    assert y1.ndim == 1 and y0.ndim==1
    if paired:
        assert len(y1) == len(y0)

    # random generator
    if random_seed is None:
        rg = np.random.RandomState()
    elif isinstance(random_seed, np.random.RandomState):
        rg = random_seed
    else:
        rg = np.random.RandomState(random_seed)

    # calculate original effect size
    d = statistic(y1) - statistic(y0)    # d: difference between statistics describing 2 data sets 

    # calculate permuted effect sizes
    perm_ds = np.empty((n_perm, *d.shape))    # to hold d calculated with permuted labels
    if not paired:
        # prepare for permutations
        ys = np.concatenate([y1, y0], axis=0)
        s1 = slice(0, y1.shape[0])
        s0 = slice(y1.shape[0], ys.shape[0])
        idc = np.arange(ys.shape[0])    
        # run permutations
        for j in range_(n_perm):
            rg.shuffle(idc)
            ys_ = ys[idc]    # order data in a random order
            perm_ds[j] = statistic(ys_[s1]) - statistic(ys_[s0])
    else:
        ys = np.array([y0, y1])
        for j in range_(n_perm):
            flip = rg.choice([False, True], size=len(y1))
            ys[flip] = np.fliplr(ys[flip])
            perm_ds[j] = statistic(ys[1]) - statistic(ys[0])

    # count probability for effect size after permutation
    # to reach equal or greater magnitude than original effect
    if alternative == 'greater':
        return (perm_ds >= d).mean()
    elif alternative == 'less':
        return (perm_ds <= d).mean()
    else:
        return (np.abs(perm_ds) >= np.abs(d)).mean()


# use numba if available for faster version
try:
    import numba
    @numba.njit
    def permutation_p_fast(y1, y0, alternative='greater', paired=False, n_perm=num_permutations, random_seed=0):
        """
        calculates p-value for mean(y1) > mean(y0) using permutation test
        notes:
        -   does not handle NaNs
        -   fixed to be one-tailed test with y1 > y0
        """
        assert y1.ndim == 1 and y0.ndim==1
        if paired:
            assert len(y1) == len(y0)

        # calculate original effect size
        d = y1.mean() - y0.mean()    # d: difference between statistics describing 2 data sets 

        perm_ds = np.empty(n_perm)    # to hold d calculated with permuted labels
        if random_seed is not None:
            np.random.seed(random_seed)

        if not paired:
            # prepare for permutations
            ys = np.concatenate((y1, y0))
            s1 = slice(0, len(y1))
            s0 = slice(len(y1), None)
            idc = np.arange(len(ys))
            # run permutations
            for j in range(n_perm):
                np.random.shuffle(idc)
                ys_ = ys[idc]    # order data in a random order
                perm_ds[j] = ys_[s1].mean() - ys_[s0].mean()
        else:
            ds = y1 - y0
            for j in range(n_perm):
                ds[np.random.random(len(ds)) > 0.5] *= -1
                perm_ds[j] = ds.mean()


        # count probability for effect size after permutation
        # to reach equal or greater magnitude than original effect
        if alternative == 'greater':
            return (perm_ds >= d).mean()
        elif alternative == 'less':
            return (perm_ds <= d).mean()
        else:
            return (np.abs(perm_ds) >= np.abs(d)).mean()

except ImportError:
    permutation_p_fast = permutation_p
    
    
stats_fun_table = {
    # Name: (n-sample, p_value_fun)
    'Binomial': (1, st.binom_test),
    'Mann-Whitney U': (2, lambda *a, **kw: st.mannwhitneyu(*a, **kw)[1]),
    'Permutation': (2, permutation_p_fast),
    'Kruskal-Wallis': (0, lambda *a, **kw: st.kruskal(*a, **kw)[1]),
    'Wilcoxon signed-rank': (2, lambda *a, **kw: st.wilcoxon(*a, **kw)[1]),
    'Paired permutation': (2, functools.partial(permutation_p_fast, paired=True)),
}


def multiple_comparisons_with_correction(
        comparisons, data, x, y, alternative='two-sided', stats_kind='', verbose=True,
        groupby=None, groups=None, alpha=0.05, correction_method='fdr_tsbky'):
    """
    correction_method: use fdr_tsbky (test statistics independent or positive dependent) or fdr_by (any dependence, but less power)
    """
    try:
        nsample, stats_fun = stats_fun_table[stats_kind]
    except KeyError:
        raise KeyError('unknown test! should be one of ', list(stats_fun_table.keys()))
    if nsample == 0 and alternative is not None:
        print(f'note: alternative {alternative} will not be used for {stats_kind} test')
            
    if groupby is not None:
        bygroup_data = dict(list(data.groupby(groupby)))
        if groups is None:
            groups = sorted(set(data[groupby].values))
        bygroup_data = {k: bygroup_data[k] for k in groups}
        grouped = True
    else:
        bygroup_data = {None: data}
        groups = (None,)
        grouped = False

    if hasattr(alternative, '__len__') and not isinstance(alternative, str):
        assert len(alternative) == len(groups)
    else:
        alternative = [alternative] * len(groups)
        
    pvals = []
    pval_keys = []
    pval_counts = []
    for i, group in enumerate(groups):
        group_data = bygroup_data[group]
        for j, comparison in enumerate(comparisons):
            comparison = tuple(comparison)
            alt = alternative[i]
            
            # collect values for comparison conditions
            cond_values = []
            cond_counts = []
            if nsample == 1:
                for y_ in y:
                    v = group_data[group_data[x]==comparison[0]][y_].values
                    v = v[~np.isnan(v)]
                    cond_values.append(v)
                cond_counts.append(len(v))
            else:
                for c in comparison:
                    v = group_data[group_data[x]==c][y].values
                    v = v[~np.isnan(v)]
                    cond_values.append(v)
                    cond_counts.append(len(v))
            
            # get p value
            if stats_kind == 'Binomial':
                pval = stats_fun(cond_values[0].sum(), n=cond_values[1].sum(), p=comparison[1], alternative=alt)
            elif nsample == 0:
                pval = stats_fun(*cond_values)
            else:
                if stats_kind in ('Permutation', 'Paired permutation'):
                    # set deterministic random seed so that results are reproducible
                    # don't set constant value for good practice
                    random_seed = data_to_random_seed(*cond_values)
                    pval = stats_fun(*cond_values, alternative=alt, random_seed=random_seed)
                else:
                    pval = stats_fun(*cond_values, alternative=alt)
            
            # collect p-values and annotation
            pvals.append(pval)
            pval_keys.append((group, comparison, alt))
            pval_counts.append(cond_counts)
                
    # multiple-comparison correction
    if len(pvals) > 1:
        mwu_reject, mwu_apvals = multipletests(pvals, alpha=alpha, method=correction_method)[:2]
    else:
        mwu_apvals = np.array(pvals)
        mwu_reject = mwu_apvals < alpha

    pval_cols = ('Group', 'Comparison', 'Alternative', 'P-value', 'Uncorrected p-value')
    pval_df = {k: [] for k in pval_cols}
    if verbose:
        print(stats_kind, 'test,', end=' ')
        if len(pvals) > 1:
            print(len(pvals), 'tests:')
        else:
            print('1 test, not doing correction')
        if verbose == 2: print(mwu_reject.sum(), 'tests showing significance')
        msg_rows = []
    for key, p, ap, reject, counts in zip(pval_keys, pvals, mwu_apvals, mwu_reject, pval_counts):
        pval_df['Group'].append(key[0])
        pval_df['Comparison'].append(key[1])
        pval_df['Alternative'].append(key[2])
        pval_df['P-value'].append(ap)
        pval_df['Uncorrected p-value'].append(p)
        if verbose:
            alternative_code = {'greater': '>', 'less': '<'}.get(key[2], '<>')
            if verbose == 2 and not reject: continue
            msg = []
            if grouped:
                msg.append(f'{groupby} = {str(key[0])},')
            comparison = key[1]
            if nsample == 0:
                msg.append(f'{len(comparison)} conditions:')
            elif nsample == 1:
                msg.append(f'{comparison[0]} ({counts[0]}) {alternative_code} {comparison[1]}:')
            elif nsample == 2 and stats_kind in ('Wilcoxon signed-rank', 'Paired permutation'):
                msg.append(f'{comparison[0]} {alternative_code} {comparison[1]} ({counts[0]}):')
            else:
                msg.append(f' {alternative_code} '.join(f'{c} ({ct})' for c, ct in zip(comparison, counts)) + ':')
            msg.append((' ', '*')[int(reject)] + f'p = {p2str(ap):<6} (uncorrected p = {p2str(p)})')
            msg_rows.append(msg)
    if verbose:
        col_lens = [max(list(map(len, col))) for col in zip(*msg_rows)]
        for msg in msg_rows:
            print('  '.join(s.ljust(l) for s, l in zip(msg, col_lens)))
    
    pval_df = pd.DataFrame.from_dict(pval_df).set_index(['Group', 'Comparison', 'Alternative'])
    pval_df.caption = f'{stats_kind} test, {len(pvals)} tests, {correction_method} correction (alpha = {alpha})'
    return pval_df


def p2str(p):
    if p == 0:
        return '0'
    elif p < 1e-3:
        return '%.0fe%d' % tuple(map(int, ('%.0e' % p).split('e')))
    else:
        return '%.3f' % p


def multiplot(
        data, x, y, palette, hue=None,
        order=None, hue_order=None, xlabels=None, split=None,
        swarm_kw=None, violin_kw=None, scatter_kw=None, vlines_kw=None,
        ci=95, ci_width=0.4,
        jitter_y=False, random_seed=0, ax=None, warn_swarm=True):
    if xlabels is not None:
        assert order is not None and len(xlabels) == len(order), 'xlabels must match order'
    palette_light = {k: brighten_color(desaturate_color(v)) for k, v in palette.items()}
    swarm_kw_ = dict(palette=palette_light, s=1.5, zorder=0, split=split, clip_on=False)
    violin_kw_ = dict(palette=palette, linewidth=.5, inner=None, cut=0, zorder=1, split=split, clip_on=False)
    scatter_kw_ = dict(s=7, zorder=2, clip_on=False)
    vlines_kw_ = dict(lw=1, zorder=2, clip_on=False)
    if swarm_kw is not None: swarm_kw_.update(swarm_kw)
    if violin_kw is not None: violin_kw_.update(violin_kw)
    if scatter_kw is not None: scatter_kw_.update(scatter_kw)
    if vlines_kw is not None: vlines_kw_.update(vlines_kw)
    if ax is None:
        ax = plt.gca()    
    if jitter_y > 0:
        data = data.copy()
        y_swarm = y+' jittered'
        data[y_swarm] = data[y] +\
            np.random.RandomState(random_seed).uniform(-jitter_y, jitter_y, size=len(data))    # for swarm
    else:
        y_swarm = y
        
    # plot swarm
    with warnings.catch_warnings():
        # swarmplot warns about no space to place points;
        # ignore as requested
        warnings.simplefilter(['ignore', 'default'][warn_swarm], category=UserWarning)

        swarmplot2(data=data, x=x, y=y_swarm, hue=hue, order=order, hue_order=hue_order, **swarm_kw_)

    # plot violin
    sns.violinplot(data=data, x=x, y=y, hue=hue, order=order, hue_order=hue_order, **violin_kw_)
    
    # plot center and CI
    for i, cat in enumerate(order):
        if hue is None:
            values = data[data[x]==cat][y].values
            c = palette[cat]
            # center
            mu = values.mean()
            ax.scatter(i, mu, color=c, **scatter_kw_)
            # CI
            boots = sns.algorithms.bootstrap(values, seed=random_seed)
            ci_ = sns.utils.ci(boots, which=ci)
            ax.vlines(i, *ci_, color=c, **vlines_kw_)
        else:
            df_ = data[data[x]==cat]
            for j, cat2 in enumerate(hue_order):
                values = df_[df_[hue]==cat2][y].values
                c = palette[cat2]
                mu = values.mean()
                j = i + (j - (len(hue_order)-1)/2) * ci_width / (len(hue_order) - 1)
                ax.scatter(j, mu, color=c, **scatter_kw_)
                boots = sns.algorithms.bootstrap(values, seed=random_seed)
                ci = sns.utils.ci(boots, which=95)
                ax.vlines(j, *ci, color=c, **vlines_kw_)

    # style violin
    violins = np.array([o for o in ax.get_children() if isinstance(o, mpl.collections.PolyCollection)])
    for i, o in enumerate(violins):
        o.set_edgecolor(o.get_facecolor())
        o.set_facecolor('none')
        o.set_clip_on(violin_kw_.get('clip_on', False))
        
    ax.tick_params(axis='x', which='major', pad=2, length=0)
    ax.spines['bottom'].set_visible(False)
    ax.set_xlabel('')
    if xlabels is not None:
        ax.set_xticklabels(xlabels)

    return ax


def plot_pvals_pairwise(
        pvals_df, order, hue_order=None, top_pairs=None, top_y0=None, bott_pairs=None, bott_y0=None,
        d=5, ang=10, width=0.5, fontsize=5, ticklen=None, label_ns=True, ax=None):
    if ax is None:
        ax = plt.gca()
    order = list(order)
    if hue_order is not None:
        hue_order = list(hue_order)
        shift_hue = lambda x: (x - (len(hue_order)-1)/2) * width / (len(hue_order) - 1)
    ang = ang/180*np.pi
        
    xrange = np.ptp(ax.get_xlim())
    yl = ax.get_ylim()
    yrange = np.ptp(yl)
    yxratio = yrange / xrange
    
    height = ax.get_window_extent().transformed(ax.figure.dpi_scale_trans.inverted()).height   # unit: inches
    d = d / 72  * (yrange / height)
    dystar = fontsize / 3 / 72 * (yrange / height)    # unit: data; adjust height of '*' down relative to fontsize in pt
    if ticklen is None:
        ticklen = d/5 if hue_order is None else 0
    
    for i, (pairs, y0) in enumerate(((bott_pairs, bott_y0), (top_pairs, top_y0))):
        if pairs is None:
            continue
        if y0 is None:
            y0 = yl[i]
        sign = 2*i-1
        for pair in pairs:
            if isinstance(pair[1], str):
                cat0, cat1 = pair[:2]
                group = None
            else:
                assert len(pair[1]) == 2
                assert hue_order is not None
                group, (cat0, cat1) = pair[:2]
            incr = 0 if len(pair) < 3 else pair[2]
            dx0 = 0 if len(pair) < 4 else pair[3]
            dx1 = 0 if len(pair) < 5 else pair[4]
            y0 += d * incr * sign
            
            if group is None:
                m = pvals_df.index.get_loc_level((cat0, cat1), level='Comparison')[0]
            else:
                m = pvals_df.index.get_loc_level((group, (cat0, cat1)), level=('Group', 'Comparison'))[0]
            pval_row = pvals_df[m]
            assert len(pval_row) == 1, f'did not find unique row for {cat0} vs {cat1}; found {pval_row.values}'
            pval = pval_row['P-value'].item()
            alt = pval_row.index.get_level_values('Alternative').item()
            if group is None:
                loc0 = order.index(cat0)+dx0
                loc1 = order.index(cat1)+dx1
            else:
                loc = order.index(group)
                loc0 = loc + shift_hue(hue_order.index(cat0))
                loc1 = loc + shift_hue(hue_order.index(cat1))
            y1 = y0 + abs(loc1-loc0) * np.tan(ang) * yxratio * dict(greater=-1, less=1).get(alt, 0)
            if pval < 0.01: s = '**'
            elif pval < 0.05: s = '*'
            else: s = 'n.s.'
            c = 'k' if s!='n.s.' else 'lightgray'
            ax.plot(
                [loc0,loc0,loc1,loc1],
                [y0-ticklen*sign,y0,y1,y1-ticklen*sign],
                color=c,
                lw=.75, clip_on=False)
            if s=='n.s.' and not label_ns:
                continue
            ax.text(
                np.mean([loc0, loc1]),
                (y0+y1)/2-(0,dystar)[(s!='n.s.')&(sign>0)],
                s, color=c,
                ha='center', va=('top', 'bottom')[i], fontsize=fontsize)


class _SwarmPlotterWithSplit(_SwarmPlotter):
    def __init__(self, x, y, hue, data, order, hue_order,
                 dodge, split, orient, color, palette):
        super().__init__(
            x, y, hue, data, order, hue_order,
            dodge, orient, color, palette)

        # Set object attributes
        self.width = .8
        if split and self.hue_names is not None and len(self.hue_names) != 2:
            msg = "There must be exactly two hue levels to use `split`.'"
            raise ValueError(msg)
        self.split = split
        if split and dodge:
            warnings.warn('Both split and dodge are true; will use split instead of dodge')
            self.dodge = False

    def position_candidates(self, xy_i, neighbors, d, split_left):
        """Return a list of (x, y) coordinates that might be valid."""
        candidates = [xy_i]
        x_i, y_i = xy_i
        left_first = True
        for x_j, y_j in neighbors:
            dy = y_i - y_j
            dx = np.sqrt(max(d ** 2 - dy ** 2, 0)) * 1.05
            cl, cr = (x_j - dx, y_i), (x_j + dx, y_i)
            if self.split:
                new_candidates = [cl] if split_left else [cr]
            elif left_first:
                new_candidates = [cl, cr]
            else:
                new_candidates = [cr, cl]
            candidates.extend(new_candidates)
            left_first = not left_first
        return np.array(candidates)

    def beeswarm(self, orig_xy, d, split_left):
        """Adjust x position of points to avoid overlaps."""
        # In this method, ``x`` is always the categorical axis
        # Center of the swarm, in point coordinates
        midline = orig_xy[0, 0]
        if self.split:
            split_offset = (.5 - split_left) * d
            midline = midline + split_offset
            orig_xy = orig_xy + [split_offset, 0]

        # Start the swarm with the first point
        swarm = [orig_xy[0]]

        # Loop over the remaining points
        for xy_i in orig_xy[1:]:

            # Find the points in the swarm that could possibly
            # overlap with the point we are currently placing
            neighbors = self.could_overlap(xy_i, swarm, d)

            # Find positions that would be valid individually
            # with respect to each of the swarm neighbors
            candidates = self.position_candidates(xy_i, neighbors, d, split_left)

            # Sort candidates by their centrality
            offsets = np.abs(candidates[:, 0] - midline)
            candidates = candidates[np.argsort(offsets)]

            # Find the first candidate that does not overlap any neighbours
            new_xy_i = self.first_non_overlapping_candidate(candidates,
                                                            neighbors, d)

            # Place it into the swarm
            swarm.append(new_xy_i)

        return np.array(swarm)

    def swarm_points(self, ax, points, center, width, s, hue_index, **kws):
        """Find new positions on the categorical axis for each point."""
        # Convert from point size (area) to diameter
        default_lw = mpl.rcParams["patch.linewidth"]
        lw = kws.get("linewidth", kws.get("lw", default_lw))
        dpi = ax.figure.dpi
        d = (np.sqrt(s) + lw) * (dpi / 72)

        # Transform the data coordinates to point coordinates.
        # We'll figure out the swarm positions in the latter
        # and then convert back to data coordinates and replot
        orig_xy = ax.transData.transform(points.get_offsets())

        # Order the variables so that x is the categorical axis
        if self.orient == "h":
            orig_xy = orig_xy[:, [1, 0]]

        # Do the beeswarm in point coordinates
        new_xy = self.beeswarm(orig_xy, d, (hue_index+1) % 2)

        # Transform the point coordinates back to data coordinates
        if self.orient == "h":
            new_xy = new_xy[:, [1, 0]]
        new_x, new_y = ax.transData.inverted().transform(new_xy).T

        # Add gutters
        if self.orient == "v":
            self.add_gutters(new_x, center, width)
        else:
            self.add_gutters(new_y, center, width)

        # Reposition the points so they do not overlap
        points.set_offsets(np.c_[new_x, new_y])

    def draw_swarmplot(self, ax, kws):
        """Plot the data."""
        s = kws.pop("s")
        centers = []
        swarms = []
        hue_level_indices = []
        palette = np.asarray(self.colors)

        # Set the categorical axes limits here for the swarm math
        if self.orient == "v":
            ax.set_xlim(-.5, len(self.plot_data) - .5)
        else:
            ax.set_ylim(-.5, len(self.plot_data) - .5)

        # Plot each swarm
        for i, group_data in enumerate(self.plot_data):
            if self.plot_hues is None or not (self.dodge or self.split):
                width = self.width

                if self.hue_names is None:
                    hue_mask = np.ones(group_data.size, bool)
                else:
                    hue_mask = np.array([h in self.hue_names
                                         for h in self.plot_hues[i]], bool)
                    # Broken on older numpys
                    # hue_mask = np.in1d(self.plot_hues[i], self.hue_names)

                swarm_data = np.asarray(group_data[hue_mask])
                point_colors = np.asarray(self.point_colors[i][hue_mask])

                # Sort the points for the beeswarm algorithm
                sorter = np.argsort(swarm_data)
                swarm_data = swarm_data[sorter]
                point_colors = point_colors[sorter]

                # Plot the points in centered positions
                cat_pos = np.ones(swarm_data.size) * i
                kws.update(c=palette[point_colors])
                if self.orient == "v":
                    points = ax.scatter(cat_pos, swarm_data, s=s, **kws)
                else:
                    points = ax.scatter(swarm_data, cat_pos, s=s, **kws)

                centers.append(i)
                swarms.append(points)
                hue_level_indices.append(i)

            else:
                offsets = self.hue_offsets
                width = self.nested_width

                for j, hue_level in enumerate(self.hue_names):
                    hue_mask = self.plot_hues[i] == hue_level
                    swarm_data = np.asarray(group_data[hue_mask])
                    point_colors = np.asarray(self.point_colors[i][hue_mask])

                    # Sort the points for the beeswarm algorithm
                    sorter = np.argsort(swarm_data)
                    swarm_data = swarm_data[sorter]
                    point_colors = point_colors[sorter]

                    # Plot the points in centered positions
                    center = i + offsets[j]
                    cat_pos = np.ones(swarm_data.size) * center
                    kws.update(c=palette[point_colors])
                    if self.orient == "v":
                        points = ax.scatter(cat_pos, swarm_data, s=s, **kws)
                    else:
                        points = ax.scatter(swarm_data, cat_pos, s=s, **kws)

                    centers.append(center)
                    swarms.append(points)
                    hue_level_indices.append(j)

        # Autoscale the valus axis to set the data/axes transforms properly
        ax.autoscale_view(scalex=self.orient == "h", scaley=self.orient == "v")

        # Update the position of each point on the categorical axis
        # Do this after plotting so that the numerical axis limits are correct
        for center, swarm, hue_index in zip(centers, swarms, hue_level_indices):
            if swarm.get_offsets().size:
                self.swarm_points(ax, swarm, center, width, s, hue_index, **kws)


def swarmplot2(
    *,
    x=None, y=None,
    hue=None, data=None,
    order=None, hue_order=None,
    dodge=False, split=False, orient=None, color=None, palette=None,
    size=5, edgecolor="gray", linewidth=0, ax=None,
    **kwargs
):

    plotter = _SwarmPlotterWithSplit(
        x, y, hue, data, order, hue_order,
        dodge, split, orient, color, palette)
    if ax is None:
        ax = plt.gca()

    kwargs.setdefault("zorder", 3)
    size = kwargs.get("s", size)
    if linewidth is None:
        linewidth = size / 10
    if edgecolor == "gray":
        edgecolor = plotter.gray
    kwargs.update(dict(s=size ** 2,
                       edgecolor=edgecolor,
                       linewidth=linewidth))

    plotter.plot(ax, kwargs)
    return ax


def make_canvas(ims, nrows=None, ncols=None, aspect_ratio=1,
                margin_colors=None, imsize=128, margin=10, cmap='gray'):
    imsize = int(imsize)
    margin = int(margin)
    if margin_colors is not None:
        assert len(ims) == len(margin_colors)
    if ncols is None:
        if nrows is None:
            ncols = int(np.ceil(np.sqrt(len(ims) * aspect_ratio)))
        else:
            ncols = int(np.ceil(len(ims) / nrows))
    nrows = int(np.ceil(len(ims) / ncols))
    size = imsize + margin
    w = margin + size * ncols
    h = margin + size * nrows
    canvas = np.ones((h, w, 3))
    for i, im in enumerate(ims):
        ih = i // ncols
        iw = i % ncols
        if len(im.shape) > 2 and im.shape[-1] == 4:
            im = im[..., :3]
        if im.shape[:2] != (imsize, imsize):
            if im.dtype != np.uint8:
                im = np.round(im*255).astype(np.uint8)
            im = np.array(Image.fromarray(im).resize((imsize, imsize), Image.ANTIALIAS))
        if im.dtype == np.uint8:
            im = im.astype(float) / 255
        if len(im.shape) == 2:
            im = get_cmap(cmap)(im)[..., :-1]
        if margin_colors is not None:
            canvas[size * ih:size * (ih + 1) + margin, size * iw:size * (iw + 1) + margin] = margin_colors[i]
        canvas[margin + size * ih:margin + size * ih + imsize, margin + size * iw:margin + size * iw + imsize] = im
    return canvas


def plot_faceness_to_resp(evo_names, evo_data, rating_df, cat_df, evo_annot_df):
    n = len(evo_names)
    c = ncols = min(n,8)
    r = int(np.ceil(n/c))

    fig, axs = plt.subplots(r, c, figsize=(.975*c, (.75,1.05)[r>1]*r), dpi=300)
    plt.subplots_adjust(hspace=0.6, wspace=0.35)
    axs = axs.ravel()
    for ax in axs[n:]:
        ax.set_axis_off()
    axs[-c].set_xlabel(f"'Faceness' rating", labelpad=0)
    
    for i, (evo_name, ax) in enumerate(zip(evo_names, axs)):
        dat = evo_data[evo_name]
        uids = dat['uids']
        resp_df = dat['resp_df']
        mask_ref_rated = dat['mask_ref_rated']
        mask_last_gen = dat['mask_last_gen']
        
        # ref images
        x0 = rating_df.loc[uids[mask_ref_rated]]

        get_color = lambda x: palette3.get(x, 'tab:blue')
        c = list(map(palette3.get, cat_df.loc[x0.index]['Category']))
        x0 = x0['Faceness rating'].values.ravel()
        y0 = resp_df.loc[uids[mask_ref_rated]].values.ravel()
        ax.scatter(x0, y0, c=c, s=3, clip_on=False)

        # evo image
        x1, dx1 = rating_df.loc[[evo_name]][['Faceness rating', 'Faceness rating SEM']].values.ravel()
        y1, dy1 = np.mean(resp_df[mask_last_gen]), st.sem(resp_df[mask_last_gen])
        
        get_color = lambda x: palette3.get(x, colors1[0])
        c = get_color(cat_df.loc[evo_name]['Category'])
        ax.hlines(y1, x1-dx1, x1+dx1, color=c, lw=.75)
        ax.vlines(x1, y1-dy1, y1+dy1, color=c, lw=.75)
        ax.scatter(x1, y1, c=[c], s=10, marker='s', clip_on=False)

        a, b = np.min([y0.min(), y1.min()]), np.max([y0.max(), y1.min()])
        res = (25, 50)[(b-a)>100]
        a = np.floor(a / res) * res
        b = np.ceil(b / res) * res
        ax.set_ylim(a, b)
        ax.set_yticks([a, b])
        ax.set_yticks(np.arange(a, b, res), minor=True)
        ax.tick_params(axis='y', which='major', pad=1)
        ax.set_xticks([1, 5])
        ax.set_xticks([2, 3, 4], minor=True)
        ax.spines['bottom'].set_bounds(1, 5)
        ax.set_xlim(.75, 5.25)
        row = evo_annot_df.loc[evo_name]
        txt = f"FSI = {row['FSI']:.2f}, r = {st.pearsonr(x0, y0)[0]:.2f}"
        ax.text(0, 1.1, txt, ha='left', va='bottom', fontsize=5, color='gray', transform=ax.transAxes)
        if i % ncols == 0:
            ax.set_ylabel('Firing rate, spks/s', labelpad=0)


def custom_scatter(x, y, s, **kwargs):
    kwargs.setdefault('c', 'k')
    kwargs.setdefault('ec', 'none')
    kwargs.setdefault('verbose', True)
    verbose = kwargs['verbose']
    del kwargs['verbose']
    
    fig = plt.figure(figsize=(1.2,1.2))
    ax = plt.gca()

    ax.scatter(x, y, s=s, **kwargs)
    for sp in ax.spines.values():
        sp.set_visible(False)
    ax.set_xticks([])
    ax.set_yticks([])
    xl = np.array(ax.get_xlim())
    yl = np.array(ax.get_ylim())
    ax.hlines(0, *xl, lw=.5, ls='-', color='gray', zorder=2)
    ax.vlines(0, *yl, lw=.5, ls='-', color='gray', zorder=2)
    for _ in np.arange(np.ceil(xl[0]/.2)*.2, (np.floor(xl[1]/.2)+1)*.2, .2):
        if not _: continue
        ax.vlines(_, 0, .025, lw=.5, ls='-', color='gray', zorder=2)
    for _ in np.arange(np.ceil(yl[0]/.25)*.25, (np.floor(yl[1]/.25)+1)*.25, .25):
        if not _: continue
        ax.hlines(_, 0, .04, lw=.5, ls='-', color='gray', zorder=2)

    # xl += .1
    yl += .1
    ax.text(-.01,-.025, '0', ha='right', va='top', transform=ax.transData, color='gray')
    ax.plot([xl[0], xl[0]+.2], [yl[0]]*2, c='k', zorder=2)
    ax.text(xl[0]+.05, yl[0]-.05, '0.2', ha='center', va='top', transform=ax.transData)
    ax.text(.5, -.01, 'Face selectivity index', ha='center', va='top', transform=ax.transAxes)
    ax.plot([xl[0]]*2, [yl[0], yl[0]+.25], c='k')
    ax.text(xl[0]-.025, yl[0]+.125, '0.25', ha='center', va='bottom', transform=ax.transData,
         rotation=90, rotation_mode='anchor')
    ax.text(-.05, 0.5, "FR correlation\nwith 'faceness'", ha='center', va='bottom', transform=ax.transAxes,
         rotation=90, rotation_mode='anchor')

    m = np.isfinite(x) & np.isfinite(y)
    r, p = st.pearsonr(x[m], y[m])
    ax.text(.0, .95, f'r = {r:.2f}', ha='left', va='top', transform=ax.transAxes, fontsize=6)
    if verbose:
        print(f'correlation r = {r:.2f}, n = {m.sum()}, p = {p:.3f} = {p:.1e}')
    return fig, m



def plot_metric_to_fsi(
        data, x, y, palette, hue='Region', hue_order=None,
        xlim=(-1, 2), xerr_type=None, yerr_type=None,
        kws=None, text_kws=None, ax=None):
    if ax is None:
        fig = plt.figure(figsize=(1.75,1.5))
        ax = plt.gca()
    else:
        fig = plt.gcf()
    figheight = fig.get_figheight()

    if hue is None:
        data_splits = {'': data}
    else:
        data_splits = dict(tuple(data.groupby(hue)))

    if palette is None:
        palette = {'': 'k'}
    
    kw_ = dict(elinewidth=0.75, lw=0, marker='.', markersize=1, alpha=0.5)
    if kws is not None:
        kw_.update(kws)
    text_kw_ = dict(ha='left', fontsize=5)
    if text_kws is not None:
        text_kw_.update(text_kws)
    
    ytext = 1
    if hue_order is None:
        hue_order = data_splits.keys()
    for i, level in enumerate(hue_order):
        data_split = data_splits[level]
        m = ~(np.isnan(data_split[x].values) | np.isnan(data_split[y].values))
        data_split = data_split[m]
        xvals = data_split[x].values
        yvals = data_split[y].values
        xerrs = None if xerr_type is None else data_split[x + ' ' + xerr_type].values
        yerrs = None if yerr_type is None else data_split[y + ' ' + yerr_type].values
    
        color = palette.get(level, plt.get_cmap('tab10')(i%10/10))
        ax.errorbar(xvals, yvals, yerrs, xerrs, color=color, ecolor=color, **kw_)

        k, b, _, pl = st.linregress(xvals, yvals)[:4]
        xl = np.array([xvals.min(), xvals.max()])
        ax.plot(xl, b + k * xl, zorder=2, color=color)
        c, pc = st.pearsonr(xvals, yvals)
        pc /= 2    # two tailed -> one tailed
        print(
            (f'{level}: ' if hue else '') 
            + 'N = %d; correlation: %.2f, p-value = %6.4f = %.0e' % (len(xvals), c, pc, pc))
        txt = ax.text(
            .01, ytext,
            (f'{level}: ' if hue else '')+ f'r = {c:.2f}, p = {p2str(pc)}',
            va='bottom', transform=ax.transAxes, color=color, **text_kw_)
        ytext += txt.get_fontsize()*1.2 / 72 /figheight
    
    ax.set_xlim(xlim)
    ax.set_xlabel(x)
    ax.set_ylabel(y)
